/*
 * This file is part of reference hash utilities
 *
 * Copyright (C) 2008-2009 Nokia Corporation and/or its subsidiary(-ies).
 *
 * Contact: Markku Kylänpää <ext-markku.kylanpaa@nokia.com>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 *
 * Authors: Markku Kylänpää 2008-2009
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <sys/types.h>
#include <sys/stat.h>

/*******************************************************************************
 * DEFINES
 ******************************************************************************/

/* Maximum character length for a hashlist entry */
#define LINE_MAX 1024

/* DigSig securityfs entry to write reference hashes */
#define DIGSIG_FILE "/sys/kernel/security/digsig/hashlist"

/*******************************************************************************
 * VARIABLES
 ******************************************************************************/

/* Test flag allows the change of target file. So this must be variable. */
char *digsig_file = DIGSIG_FILE;

/* Mount point */
char mnt_point[LINE_MAX];

/*******************************************************************************
 * TYPES
 ******************************************************************************/

/* Command line options */
struct parameters {
	int verbose;
	int disable;
};

/*******************************************************************************
 * MODULE STATIC FUNCTIONS
 ******************************************************************************/

/*
 * Print usage information.
 */
static void
usage(const char *filename)
{
	printf("Usage: %s [[-n]]|-d|-c filename|-h|-v] "
	       "filename\n",
	       filename);
	printf("  -c filename Securefs/sysfs hashfile to write\n");
	printf("     Default is /sys/kernel/security/digsig/hashlist\n");
	printf("  -m <mount point path> default is /\n");
        printf("  -n Do not write entries - use /dev/null\n");
        printf("  -v Verbose output\n");
        printf("  -h Print usage information\n");
}

/*
 * Parse command options
 */
static const char*
parse_args(int argc, char **argv, struct parameters *ops)
{
	int argno = 0;
	const char *filename = NULL;

	if (argc < 2) {
		usage(argv[0]);
		exit(0);
	}

	if (ops) {
		ops->verbose = 0;
		ops->disable = 0;
	} else {
		return NULL;
	}

	argno++;
	while (argno < argc) {
		if (strcmp(argv[argno], "-c") == 0) {
			if (++argno < argc) {
				digsig_file = argv[argno];
			} else {
				usage(argv[0]);
				exit(0);
			}
		} else if (strcmp(argv[argno], "-n") == 0) {
			ops->disable = 1;
		} else if (strcmp(argv[argno], "-h") == 0) {
			usage(argv[0]);
			exit(0);
		} else if (strcmp(argv[argno], "-v") == 0) {
			ops->verbose = 1;
		} else if (strcmp(argv[argno], "-m") == 0) {
			strncpy(mnt_point, argv[++argno], LINE_MAX);
			mnt_point[LINE_MAX - 1] = '\0';
		} else {
			filename = argv[argno];
		}
		argno++;
	}
	return filename;
}

/*
 * Convert hexadecimal number (character) to a binary value
 */
static int
hex2bin(int c)
{
	c = tolower(c);
	if ((c >= 'a') && (c <= 'f')) {
		return c -'a' + 10;
	}
	if ((c >= '0') && (c <= '9')) {
		return c - '0';
	}
	return -1;
}

/*
 * Get device identifier of the filesystem.
 */
static dev_t
get_device_id(const char *filename)
{
	struct stat statbuf;

	if (stat(filename, &statbuf) != 0) {
		fprintf(stderr, "Cannot stat file %s\n", filename);
		exit(1);
	}
	return statbuf.st_dev;
}

/*
 * Read a file containing reference hashes and write data to DigSig.
 */
static void
feed_hash_entries(const char *filename, struct parameters *ops)
{
	char buffer[LINE_MAX];
	char hash[LINE_MAX];
	char fname[LINE_MAX];
	FILE *fp;
        FILE *digsigfp;
	unsigned long ino;
	dev_t device_id = 0;

	fp = fopen(filename, "r");
	if (fp == NULL) {
		printf("Cannot open hashfile %s\n", filename);
		exit(1);
	}
	while (fgets(buffer, LINE_MAX, fp) != NULL) {
		int c;
		int j;
		int len;
		char *bufptr;
		char *token;
		char *exename;
		char *pathname;
		struct stat state;
		int retval;

		if (ops->disable) {
			digsigfp = fopen("/dev/null", "w");
		} else {
			digsigfp = fopen(digsig_file, "w");
		}

		if (digsigfp == NULL) {
			printf("Cannot open DigSig file %s\n", digsig_file);
			exit(1);
		}

		bufptr = buffer;
		while ((token = strsep(&bufptr, " ")) != NULL) {
			if (strcmp(token, "S") == 0) {
				token = strsep(&bufptr, " ");
				token = strsep(&bufptr, " ");
			} else if (strcmp(token, "H") == 0) {
				token = strsep(&bufptr, " ");
				token = strsep(&bufptr, " ");
				strcpy(hash, token);
			} else if (strcmp(token, "R") == 0) {
				token = strsep(&bufptr, " ");
				len = atoi(token);
				exename = strsep(&bufptr, "\n");
				strcpy(fname, exename);
			}
		}

		if (ops->verbose) {
			printf("SHA1=%s File=%s\n", hash, fname);
		}

		pathname = malloc(strlen(mnt_point) + strlen(fname) + 2);
		strcpy(pathname, mnt_point);
		if ((pathname[strlen(pathname) - 1] == '/') ||
		    (fname[0] == '/')) {
			strcat(pathname, fname);
		} else {
			strcat(pathname, "/");
			strcat(pathname, fname);
		}
		retval = stat(pathname, &state);
		if (retval == 0) {
			ino = state.st_ino;
		} else {
			fprintf(stderr, "File not found %s (%d)\n",
				pathname, retval);
			fclose(digsigfp);
			free(pathname);
			continue;
		}

		device_id = get_device_id(pathname);
		if (ops->verbose) {
			fprintf(stderr, "Sending %s ", pathname);
		}
		c = 'a';
		fputc(c, digsigfp);
		fprintf(stderr, "Send: %c", c);
		bufptr = hash;
		for (j = 0; j < 20; j++) {
			c = hex2bin(*bufptr++);
			c = c << 4 | hex2bin(*bufptr++);
			fputc(c, digsigfp);
			fprintf(stderr, "%.2x", c);
		}

		fprintf(digsigfp, " %llu %lu\n", device_id, ino);
		fprintf(stderr, " %llu %lu\n", device_id, ino);

		fclose(digsigfp);
		if (ops->verbose) {
			fprintf(stderr, "DONE\n");
		}
		free(pathname);
	}

	fclose(fp);
}

/*******************************************************************************
 * MAIN FUNCTION
 ******************************************************************************/
int
main(int argc, char **argv)
{
	const char *filename = NULL;
	struct parameters ops;

	strcpy(mnt_point, "/");
	filename = parse_args(argc, argv, &ops); 
	if (filename) {
		feed_hash_entries(filename, &ops);
	} else {
		usage(argv[0]);
	}
	return(0);
}
