/*
 * This file is part of reference hash utilities
 *
 * Copyright (C) 1998, 1999, 2000, 2001 Free Software Foundation, Inc.
 * Copyright (C) 2004 g10 Code GmbH
 * Copyright (C) 2008-2009 Nokia Corporation and/or its subsidiary(-ies).
 *
 * Contact: Markku Kylänpää <ext-markku.kylanpaa@nokia.com>
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 *
 * refhashmake - Calculate SHA-1 checksum of ELF files. Exclude BSign section
 *               if it is found from the binary file in order to be compatible
 *               with DigSig cryptographic hash calculation algorithm.
 *
 * This software is based on sha1sum.c (see information below). The
 * original code is extended to recognize ELF files and ELF file
 * signature section. Default behavior (which is compatible with BSign)
 * is to zeroize the signature section in cryptographic hash calculation.
 * Also filename, inode and file size information is included in output. 
 *
 * sha1sum.c - print SHA-1 Message-Digest Algorithm 
 *
 * SHA-1 code is taken from gnupg 1.3.92. 
 * ELF binary analysis code is taken from readelf (part of binutils-2.16.94)
 *
 * Authors: Markku Kylänpää 2008-2009
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>
#include <errno.h>
#include <limits.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <getopt.h>
#include <elf.h>
#include "sha1.h"

/*******************************************************************************
 * DEFINES
 ******************************************************************************/

/* SHA-1 cryptographic length in bytes */
#define SHA1_LENGTH 20

/* Block size for reading data to SHA-1 calculation */
#define HBUFSIZE 4096

/* Command-line flags */
enum { RH_VERBOSE = 0,
       RH_RELATIVE = 1,
       RH_SCRIPTS = 2,
       RH_ELF = 3,
       RH_NOLINKS = 4,
       RH_ZEROIZE = 5,
       RH_NOEXEBIT = 6
};


/*******************************************************************************
 * MODULE STATIC FUNCTIONS
 ******************************************************************************/


/*
 * Pathname is typically relative to top directory, but we want to print it as
 * full pathname staring with '/' unless the flag 'relative' is set. In order
 * to ease parsing also the length of the pathname is printed before the string.
 */
static void
print_pathname(FILE *fp, const char *fname, int flags, const char *prefix)
{
	int length = strlen(fname);
	int plen = 0;
	char pstyle;
	int offset;

	if (prefix) {
		plen = strlen(prefix);
	}

	if (flags & (1 << RH_RELATIVE)) {
		offset = 1;
		pstyle = 'R';
	} else {
		offset = 0;
		pstyle = 'F';
	}

	if ((strlen(fname) > 2) && (strncmp(fname, "./", 2) == 0))  {
		if (plen) {
			fprintf(fp, "%c %d %s%s", pstyle,
				length - 1 - offset + plen, prefix,
				fname + 1 + offset);
		} else {
			fprintf(fp, "%c %d %s", pstyle,
				length - 1 - offset, fname + 1 + offset);
		}
	} else {
		if (fname[0] != '/') {
			if (plen) {
				fprintf(fp, "%c %d %s%s", pstyle,
					length + plen + 1 - offset, prefix,
					fname);
			} else {
				fprintf(fp, "%c %d %s", pstyle,
					length + 1 - offset, fname);
			}
		} else {
			if (plen) {
				fprintf(fp, "%c %d %s%s", pstyle,
					length - offset + plen, prefix,
					fname + offset);
			} else {
				fprintf(fp, "%c %d %s", pstyle, length - offset,
					fname + offset);
			}
		}
	}
}

/*
 * Print output line: [inode,] [size,] [sourceid,] SHA-1 hash, filename.
 * Information is written as one line per entry using the following
 * tag-length-value syntax. There is one letter tag that is followed by
 * a space and length field if the field value is alphaphetic. Numeric
 * values did not have length fields. Tags are:
 *
 * H n Hash value n bytes - Currently only SHA1 is used
 * R n Relative pathname n bytes
 * F n Full pathname n bytes
 * I   Inode number
 * S n Source identifier n bytes
 * L   Size of the file
 * 
 */
static void
print_output(FILE *fp, const char *fname, const unsigned char *hash, size_t len,
	     int flags, const char *sourceid, const char *prefix)
{
	struct stat statbuf;
	int i;

	if (stat(fname, &statbuf) != 0) {
		fprintf(stderr, "Cannot stat a file %s\n", fname);
		exit(1);
	}

	/* The format should include sourceid field. This is still unused */
	if (sourceid) {
		fprintf(fp, "S %d %s ", strlen(sourceid), sourceid);
	}
	
	/* Print SHA-1 hash */
	fprintf(fp, "H %d ", 2 * SHA1_LENGTH);
	for (i = 0; i < len; i++) {
		unsigned int val = (unsigned int)hash[i];
		fprintf(fp, "%02x", val);
	}
	
	/* Print filename */
	fprintf(fp, " ");
	print_pathname(fp, fname, flags, prefix);
	fprintf(fp, "\n");
}

/*
 * Read ELF header from the file stream. Return 0 if reading failed.
 * Caller should free the returned buffer. Note that this function
 * does not verify that the file really is an ELF file.
 */
static Elf32_Ehdr*
read_elf_header(FILE *fp)
{
	unsigned char *buffer;
	int size;
	int count;

	size = sizeof(Elf32_Ehdr);
	buffer = malloc(size);
	if (buffer == 0) {
		fprintf(stderr, "Malloc failed\n");
		return 0;
	}
	count = fread(buffer, 1, size, fp);
	if (count < size) {
		free(buffer);
		return 0;
	}
	return (Elf32_Ehdr*)buffer;
}

/*
 * Read num number of section headers starting from offset. Sections are
 * stored in buffer as a return value. The caller should free the buffer.
 */
static Elf32_Shdr*
read_section_headers(FILE *fp, int offset, int num)
{
	unsigned char *buffer;
	int count;
	int size = num * sizeof(Elf32_Shdr);
	if (fseek(fp, offset, SEEK_SET) < 0) {
		return 0;
	}
	buffer = malloc(size);
	if (buffer == 0) {
		return 0;
	}
	count = fread(buffer, 1, size, fp);
	if (count < size) {
		free(buffer);
		return 0;
	}
	return (Elf32_Shdr*)buffer;
}

/*
 * Find signature section from the ELF header. Return the offset of the
 * signature or -1 if there is no signature section.
 */
static int
find_signature(Elf32_Ehdr *elf_ex, Elf32_Shdr *elf_shdata, size_t *size)
{
	const unsigned int digsig_elf_sig_section = 0x80736967;

	int i = 0;

	while (i < elf_ex->e_shnum
	       && elf_shdata[i].sh_type != digsig_elf_sig_section) {
		i++;
	}

	if (i == elf_ex->e_shnum)
		return -1;
	if (elf_shdata[i].sh_type != digsig_elf_sig_section)
		return -1;
	*size = elf_shdata[i].sh_size;
	return elf_shdata[i].sh_offset;
}

/*
 * Test whether the file is an ELF file. Open the file and read ELF header.
 * Make a some sanity checks to ELF header data. Sanity checks are from DigSig.
 *
 * Return 1 if the file is ELF file and 0 otherwise.
 */
static int
is_elf_file(Elf32_Ehdr *elfhdr)
{
	int retval;

	if (memcmp(elfhdr->e_ident, ELFMAG, SELFMAG) != 0) {
		retval = 0;
		goto errclose;
	}
	if (!elfhdr->e_shoff) {
		retval = 0;
		goto errclose;
	}
	if (elfhdr->e_shentsize != sizeof(Elf32_Shdr)) {
		retval = 0;
		goto errclose;
	}
	if (elfhdr->e_shnum > 65536U / sizeof(Elf32_Shdr)) {
		retval = 0;
		goto errclose;
	}
	retval = 1;
errclose:
	return retval;
}

/*
 * Calculate SHA-1 hash for the file. If there is signature section
 * zeroize it before calculation to be compatible with DigSig kernel
 * module.
 */
static int
calculate_hash(FILE *fp, sha1_context *ctx)
{
	unsigned char buffer[HBUFSIZE];
	size_t n;
	int offset = 0;
	int low = 0;
	int up = 0;

	if (fseek(fp, 0, SEEK_SET)< 0) {
		fprintf(stderr, "cannot seek\n");
		return 0;
	}
	if (ctx->zeroize) {
		low = ctx->offset;
		up = low + ctx->size;
	}
	sha1_init(ctx);
	while ((n = fread(buffer, 1, HBUFSIZE, fp))) {
		if (ctx->zeroize) {
			if ((low < offset + n) && (offset < up)) {
				low = (offset > low) ? offset : low;
				up = (offset + n < up) ? offset + n : up;
				memset(buffer + (low - offset), 0, up - low);
			}
		}
		sha1_update(ctx, buffer, n);
		offset += n;
	}
	if (ferror(fp)) {
		fprintf(stderr, "error reading\n");
		return 0;
	}
	sha1_final(ctx);
	return 1;
}

/*
 * Print usage information.
 */
static void
usage(const char *filename)
{
	printf("Usage: %s [OPTION...] [FILE]...\n", filename);
	printf("%s calculates reference hashes from ELF binaries. "
	       "Filenames should be given as arguments.\n", filename);
	printf("Alternatively filenames can be read from a file. These ");
	printf("hashes are recalculated and verified by Linux\nkernel module ");
	printf("to authorize execution of ELF applications and libraries\n\n");
	printf("Options:\n\n");
	printf("  -a, --all            Process both scripts and ELF files\n");
	printf("  -c, --create         Use new format with tag-length-value\n");
	printf("  -f, --file           Filename(s) as an argument\n"); 
	printf("  -h, --help           Print usage information\n");
	printf("  -n  --no-links       Do not generate hashes for symlinks\n");
	printf("  -b  --no-exebit      Do not test execute file attribute\n");
	printf("  -o, --source-id      Include componet source identifier "
	       "as an argument\n");
	printf("  -p  --prefix         Add prefix argument to filename\n"); 
	printf("  -r, --relative-path  Use realtive pathnames in output\n");
	printf("  -u  --full-path      Use full pathanmes in output\n");
	printf("  -s  --scripts        Process only scripts\n");
	printf("  -e  --elf            Process only ELF files\n");
	printf("  -x, --xfiles         Extract filenames from the file given "
	       "as an argument\n");
	printf("  -z, --zeroize        Treat .signature ELF section to contain"
	       " zero value\n");
}

/*
 * Process file and if the file is either ELF application or shared library
 * calculate and print SHA-1 hash of the file (special processing for files
 * which have signature section).
 */
static int
process_file(const char *filename, int flags, const char *sourceid,
	     const char *prefix)
{
	FILE *fp;
	Elf32_Ehdr *elfhdr;
	Elf32_Shdr *elfsecs;
	sha1_context ctx;
	int offset;
	char buf[32];
	size_t n;
	int status;
	struct stat mystat;
	size_t sig_section_size;
	ctx.zeroize = ctx.offset = ctx.size = 0;
	mode_t executable = S_IXUSR | S_IXGRP | S_IXOTH;
	
	if (flags & (1 << RH_NOLINKS)) {
		status = lstat(filename, &mystat);
		if (status)
			return -1;
		switch (mystat.st_mode & S_IFMT) {
		case S_IFLNK:
			return 0;
			break;
		case S_IFDIR:
			return 0;
			break;
		case S_IFREG:
			if (!(flags & (1 << RH_NOEXEBIT)) &&
			    (mystat.st_mode & executable)) {
				fp = fopen(filename, "r");
				if (!fp) {
					if (flags & (1 << RH_VERBOSE)) {
						fprintf(stderr, "Cannot open"
							" %s\n", filename);
					}
					return -1;
				}
				if (calculate_hash(fp, &ctx)) {
					fclose(fp);
					print_output(stdout, filename, ctx.buf,
						     SHA1_LENGTH, flags,
						     sourceid, prefix);
					return 0;
				}
				fclose(fp);
			}
			break;
		default:
			break;
		}
	}

	if (!(flags & (1 << RH_NOEXEBIT))) { 
		status = lstat(filename, &mystat);
		if (status)
			return -1;
		if (S_ISREG(mystat.st_mode) && (mystat.st_mode & executable)) {
			fp = fopen(filename, "r");
			if (!fp) {
				if (flags & (1 << RH_VERBOSE)) {
					fprintf(stderr, "Cannot open"
						" %s\n", filename);
				}
				return -1;
			}
			if (calculate_hash(fp, &ctx)) {
				fclose(fp);
				print_output(stdout, filename, ctx.buf,
					     SHA1_LENGTH, flags,
					     sourceid, prefix);
				return 0;
			}
			fclose(fp);
		}
	}

	fp = fopen(filename, "r");
	if (!fp) {
		if (flags & (1 << RH_VERBOSE)) {
			fprintf(stderr, "Cannot open %s\n", filename);
		}
		return -1;
	}
	if (flags & (1 << RH_SCRIPTS)) {
	        n = fread(buf, 1, 2, fp);
		if ((n == 2) && (buf[0] == '#') && (buf[1] == '!')) {
			if (calculate_hash(fp, &ctx)) {
				fclose(fp);
				print_output(stdout, filename, ctx.buf,
					     SHA1_LENGTH, flags, sourceid,
					     prefix);
				return 0;
			}
		} else {
			if (flags & (1 << RH_ELF)) {
				if (fseek(fp, 0, SEEK_SET)) {
					fprintf(stderr, "Unable to seek %s\n",
						filename);
					fclose(fp);
					return -1;
				}
			} else {
				fclose(fp);
				return -1;
			}
		}
	}
	elfhdr = read_elf_header(fp);
	if (!elfhdr) {
		if (flags & (1 << RH_VERBOSE)) {
			fprintf(stderr, "Problem reading ELF header %s\n",
				filename);
		}
		fclose(fp);
		return -1;
	}
	if (!is_elf_file(elfhdr)) {
		if (flags & (1 << RH_VERBOSE)) {
			fprintf(stderr, "Not ELF file %s\n", filename);
		}
		fclose(fp);
		free(elfhdr);
		return -1;
	}
	elfsecs = read_section_headers(fp, elfhdr->e_shoff, elfhdr->e_shnum);
	if (!elfsecs) {
		if (flags & (1 << RH_VERBOSE)) {
			fprintf(stderr, "Problem reading ELF sections %s\n",
				filename);
		}
		fclose(fp);
		free(elfhdr);
		return -1;
	}
	offset = find_signature(elfhdr, elfsecs, &sig_section_size);
	if (offset > 0) {
		ctx.zeroize = (flags & (1 << RH_ZEROIZE)) ? 1 : 0;
		ctx.offset = offset;
		ctx.size = sig_section_size;
	}
	if (calculate_hash(fp, &ctx)) {
		fclose(fp);
		free(elfsecs);
		free(elfhdr);
		print_output(stdout, filename, ctx.buf, SHA1_LENGTH, flags,
			     sourceid, prefix);
	} else {
		if (flags & (1 << RH_VERBOSE)) {
			fprintf(stderr, "Hash calculation failed %s\n",
				filename);
		}
		fclose(fp);
		free(elfsecs);
		free(elfhdr);
		return -1;
	}
	return 0;
}

/*******************************************************************************
 * MAIN FUNCTION
 ******************************************************************************/
int 
main(int argc, char **argv)
{
	int flags = (1 << RH_ELF) | (1 << RH_SCRIPTS) | (1 << RH_RELATIVE);
	int status = 0;
	char *sourceid = NULL;
	char *prefix = NULL;
	int c;
      
	while (1) {
		static struct option long_options[] = {
			{"verbose",        no_argument, 0, 'v'},
			{"create",         no_argument, 0, 'c'},
			{"relative-path",  no_argument, 0, 'r'},
			{"full-path",      no_argument, 0, 'u'},
			{"no-links",       no_argument, 0, 'n'},
			{"source-id",      required_argument, 0, 'o'},
			{"prefix",         required_argument, 0, 'p'},
			{"scripts",        no_argument, 0, 's'},
			{"elf",            no_argument, 0, 'e'},
			{"all",            no_argument, 0, 'a'},
			{"dot",            no_argument, 0, 'd'},
			{"help",           no_argument, 0, 'h'},
			{"zeroize",        no_argument, 0, 'z'},
			{"no-exebit",      no_argument, 0, 'b'},
			{"xfiles",         required_argument, 0, 'x'},
			{"file",           required_argument, 0, 'f'},
			{0, 0, 0, 0}
		};
		int option_index = 0;
		c = getopt_long (argc, argv, "vbcruno:p:seahx:f:",
				 long_options, &option_index);
		if (c == -1)
			break;
		
		switch (c) {
		case 0:
			break;
		case 'v':
			flags |= (1 << RH_VERBOSE);
			break;
		case 'c':
			break;
		case 'r':
			flags |= (1 << RH_RELATIVE);
			break;
		case 'u':
			flags &= ~(1 << RH_RELATIVE);
			break;
		case 'n':
			flags |= (1 << RH_NOLINKS);
			break;
		case 'o':
			sourceid = malloc(strlen(optarg) + 1);
			strcpy(sourceid, optarg);
			break;
		case 'p':
			prefix = malloc(strlen(optarg) + 1);
			strcpy(prefix, optarg);
			break;
		case 's':
			flags |= (1 << RH_SCRIPTS);
			flags &= ~(1 << RH_ELF);
		        break;
		case 'e':
			flags |= (1 << RH_ELF);
			flags &= ~(1 << RH_SCRIPTS);
		        break;
		case 'a':
			flags |= (1 << RH_SCRIPTS);
			flags |= (1 << RH_ELF);
		        break;
		case 'z':
			flags |= (1 << RH_ZEROIZE);
			break;
		case 'b':
			flags |= (1 << RH_NOEXEBIT);
			break;
		case 'h':
			usage(argv[0]);
			goto terminate;
			break;
		case 'x':
		{
			char filename[LINE_MAX];
			FILE *fp = fopen(optarg, "r");
			if (fp == NULL) {
				fprintf(stderr, "Cannot open file %s\n",
					optarg);
				status = -1;
				goto terminate;
			}
			while (fgets(filename, LINE_MAX, fp) != NULL) {
				filename[strlen(filename)-1] = '\0';
				process_file(filename, flags, sourceid, prefix);
			}
			fclose(fp);
			goto terminate;
		}
		case 'f':
			process_file(optarg, flags, sourceid, prefix);
			while (optind < argc) {
				process_file(argv[optind++], flags, sourceid,
					     prefix);
			}
			goto terminate;
			break;
		case '?':
			usage(argv[0]);
			break;
		default:
			usage(argv[0]);
		}
	}

terminate:
	if (sourceid) {
		free(sourceid);
	}
	if (prefix) {
		free(prefix);
	}
	exit(status);
}

